@@ -5,11 +5,14 @@
 #
 
 from collections.abc import Sequence
-from sqlalchemy import true, false
-from sqlalchemy.sql.dml import UpdateBase
-from sqlalchemy.util.compat import string_types
-from sqlalchemy.sql.elements import ClauseElement
+from typing import List
 
+from sqlalchemy import false, true
+from sqlalchemy.sql.ddl import DDLElement
+from sqlalchemy.sql.dml import UpdateBase
+from sqlalchemy.sql.elements import ClauseElement
+from sqlalchemy.sql.roles import FromClauseRole
+from sqlalchemy.util.compat import string_types
 
 NoneType = type(None)
 
@@ -83,6 +86,18 @@ class MergeInto(UpdateBase):
         return clause
 
 
+class FilesOption:
+    """
+    Class to represent FILES option for the snowflake COPY INTO statement
+    """
+    def __init__(self, file_names: List[str]):
+        self.file_names = file_names
+
+    def __str__(self):
+        the_files = ["'" + f.replace("'", "\\'") + "'" for f in self.file_names]
+        return f"({','.join(the_files)})"
+
+
 class CopyInto(UpdateBase):
     """Copy Into Command base class, for documentation see:
     https://docs.snowflake.net/manuals/sql-reference/sql/copy-into-location.html"""
@@ -90,51 +105,88 @@ class CopyInto(UpdateBase):
     __visit_name__ = 'copy_into'
     _bind = None
 
-    def __init__(self, from_, into, formatter):
+    def __init__(self, from_, into, formatter=None):
         self.from_ = from_
         self.into = into
         self.formatter = formatter
         self.copy_options = {}
 
     def __repr__(self):
-        options = (' ' + ' '.join(["{} = {}".format(n, str(v)) for n, v in
-                                   self.copy_options.items()])) if self.copy_options else ''
-        return "COPY INTO {} FROM {} {}{}".format(self.into.__repr__(),
-                                                  self.from_.__repr__(),
-                                                  self.formatter.__repr__(),
-                                                  options)
+        """
+        repr for debugging / logging purposes only. For compilation logic, see
+        the corresponding visitor in base.py
+        """
+        return f"COPY INTO {self.into} FROM {repr(self.from_)} {repr(self.formatter)} ({self.copy_options})"
 
     def bind(self):
         return None
 
-    def overwrite(self, overwrite):
-        if not isinstance(overwrite, bool):
-            raise TypeError("Parameter overwrite should  be a boolean value")
-        self.copy_options.update({'OVERWRITE': translate_bool(overwrite)})
+    def force(self, force):
+        if not isinstance(force, bool):
+            raise TypeError("Parameter force should be a boolean value")
+        self.copy_options.update({'FORCE': translate_bool(force)})
+        return self
 
     def single(self, single_file):
         if not isinstance(single_file, bool):
             raise TypeError("Parameter single_file should  be a boolean value")
         self.copy_options.update({'SINGLE': translate_bool(single_file)})
+        return self
 
     def maxfilesize(self, max_size):
         if not isinstance(max_size, int):
             raise TypeError("Parameter max_size should be an integer value")
         self.copy_options.update({'MAX_FILE_SIZE': max_size})
+        return self
+
+    def files(self, file_names):
+        self.copy_options.update({'FILES': FilesOption(file_names)})
+        return self
+
+    def pattern(self, pattern):
+        self.copy_options.update({'PATTERN': pattern})
+        return self
 
 
 class CopyFormatter(ClauseElement):
+    """
+    Base class for Formatter specifications inside a COPY INTO statement. May also
+    be used to create a named format.
+    """
     __visit_name__ = 'copy_formatter'
 
-    def __init__(self):
-        self.options = {}
+    def __init__(self, format_name=None):
+        self.options = dict()
+        if format_name:
+            self.options["format_name"] = format_name
 
     def __repr__(self):
-        return 'FILE_FORMAT=(TYPE={}{})'.format(
-            self.file_format,
-            (' ' + ' '.join([("{} = '{}'" if isinstance(value, str) else "{} = {}").format(name, str(value))
-                             for name, value in self.options.items()])) if self.options else ""
-        )
+        """
+        repr for debugging / logging purposes only. For compilation logic, see
+        the corresponding visitor in base.py
+        """
+        return f"FILE_FORMAT=({self.options})"
+
+    @staticmethod
+    def value_repr(name, value):
+        """
+        Make a SQL-suitable representation of "value". This is called from
+        the corresponding visitor function (base.py/visit_copy_formatter())
+        - in case of a format name: return it without quotes
+        - in case of a string: enclose in quotes: "value"
+        - in case of a tuple of length 1: enclose the only element in brackets: (value)
+            Standard stringification of Python would append a trailing comma: (value,)
+            which is not correct in SQL
+        - otherwise: just convert to str as is: value
+        """
+        if name == "format_name":
+            return value
+        elif isinstance(value, str):
+            return f"'{value}'"
+        elif isinstance(value, tuple) and len(value) == 1:
+            return f"('{value[0]}')"
+        else:
+            return str(value)
 
 
 class CSVFormatter(CopyFormatter):
@@ -150,11 +202,26 @@ class CSVFormatter(CopyFormatter):
         self.options['COMPRESSION'] = comp_type
         return self
 
+    def _check_delimiter(self, delimiter, delimiter_txt):
+        """
+        Check if a delimiter is either a string of length 1 or an integer. In case of
+        a string delimiter, take into account that the actual string may be longer,
+        but still evaluate to a single character (like "\\n" or r"\n"
+        """
+        if isinstance(delimiter, NoneType):
+            return
+        if isinstance(delimiter, string_types):
+            delimiter_processed = delimiter.encode().decode("unicode_escape")
+            if len(delimiter_processed) == 1:
+                return
+        if isinstance(delimiter, int):
+            return
+        raise TypeError(
+            "{} should be a single character, that is either a string, or a number".format(delimiter_txt))
+
     def record_delimiter(self, deli_type):
         """Character that separates records in an unloaded file."""
-        if not isinstance(deli_type, (int, string_types)) \
-                or (isinstance(deli_type, string_types) and len(deli_type) != 1):
-            raise TypeError("Record delimeter should be a single character, that is either a string, or a number")
+        self._check_delimiter(deli_type, "Record delimiter")
         if isinstance(deli_type, int):
             self.options['RECORD_DELIMITER'] = hex(deli_type)
         else:
@@ -163,9 +230,7 @@ class CSVFormatter(CopyFormatter):
 
     def field_delimiter(self, deli_type):
         """Character that separates fields in an unloaded file."""
-        if not isinstance(deli_type, (int, NoneType, string_types)) \
-                or (isinstance(deli_type, string_types) and len(deli_type) != 1):
-            raise TypeError("Field delimeter should be a single character, that is either a string, or a number")
+        self._check_delimiter(deli_type, "Field delimiter")
         if isinstance(deli_type, int):
             self.options['FIELD_DELIMITER'] = hex(deli_type)
         else:
@@ -214,9 +279,7 @@ class CSVFormatter(CopyFormatter):
 
     def escape(self, esc):
         """Character used as the escape character for any field values."""
-        if not isinstance(esc, (int, NoneType, string_types)) \
-                or (isinstance(esc, string_types) and len(esc) != 1):
-            raise TypeError("Escape should be a single character, that is either a string, or a number")
+        self._check_delimiter(esc, "Escape")
         if isinstance(esc, int):
             self.options['ESCAPE'] = hex(esc)
         else:
@@ -225,10 +288,7 @@ class CSVFormatter(CopyFormatter):
 
     def escape_unenclosed_field(self, esc):
         """Single character string used as the escape character for unenclosed field values only."""
-        if not isinstance(esc, (int, NoneType, string_types)) \
-                or (isinstance(esc, string_types) and len(esc) != 1):
-            raise TypeError(
-                "Escape unenclosed field should be a single character, that is either a string, or a number")
+        self._check_delimiter(esc, "Escape unenclosed field")
         if isinstance(esc, int):
             self.options['ESCAPE_UNENCLOSED_FIELD'] = hex(esc)
         else:
@@ -243,12 +303,40 @@ class CSVFormatter(CopyFormatter):
         self.options['FIELD_OPTIONALLY_ENCLOSED_BY'] = enc
         return self
 
-    def null_if(self, null):
+    def null_if(self, null_value):
         """Copying into a table these strings will be replaced by a NULL, while copying out of Snowflake will replace
         NULL values with the first string"""
-        if not isinstance(null, Sequence):
-            raise TypeError('Parameter null should be an iterable')
-        self.options['NULL_IF'] = tuple(null)
+        if not isinstance(null_value, Sequence):
+            raise TypeError('Parameter null_value should be an iterable')
+        self.options['NULL_IF'] = tuple(null_value)
+        return self
+
+    def skip_header(self, skip_header):
+        """
+        Number of header rows to be skipped at the beginning of the file
+        """
+        if not isinstance(skip_header, int):
+            raise TypeError("skip_header  should be an int")
+        self.options['SKIP_HEADER'] = skip_header
+        return self
+
+    def trim_space(self, trim_space):
+        """
+        Remove leading or trailing white spaces
+        """
+        if not isinstance(trim_space, bool):
+            raise TypeError("trim_space should be a bool")
+        self.options['TRIM_SPACE'] = trim_space
+        return self
+
+    def error_on_column_count_mismatch(self, error_on_col_count_mismatch):
+        """
+        Generate a parsing error if the number of delimited columns (i.e. fields) in
+        an input data file does not match the number of columns in the corresponding table.
+        """
+        if not isinstance(error_on_col_count_mismatch, bool):
+            raise TypeError("skip_header  should be a bool")
+        self.options['ERROR_ON_COLUMN_COUNT_MISMATCH'] = error_on_col_count_mismatch
         return self
 
 
@@ -288,6 +376,84 @@ class PARQUETFormatter(CopyFormatter):
         self.options['SNAPPY_COMPRESSION'] = translate_bool(comp)
         return self
 
+    def compression(self, comp):
+        """
+        Set compression type
+        """
+        if not isinstance(comp, str):
+            raise TypeError("Comp should be a str value")
+        self.options['COMPRESSION'] = comp
+        return self
+
+    def binary_as_text(self, value):
+        """Enable, or disable binary as text"""
+        if not isinstance(value, bool):
+            raise TypeError("binary_as_text should be a Boolean value")
+        self.options['BINARY_AS_TEXT'] = translate_bool(value)
+        return self
+
+
+class ExternalStage(ClauseElement, FromClauseRole):
+    """External Stage descriptor"""
+    __visit_name__ = "external_stage"
+    _hide_froms = ()
+
+    @staticmethod
+    def prepare_namespace(namespace):
+        return "{}.".format(namespace) if not namespace.endswith(".") else namespace
+
+    @staticmethod
+    def prepare_path(path):
+        return "/{}".format(path) if not path.startswith("/") else path
+
+    def __init__(self, name, path=None, namespace=None, file_format=None):
+        self.name = name
+        self.path = self.prepare_path(path) if path else ""
+        self.namespace = self.prepare_namespace(namespace) if namespace else ""
+        self.file_format = file_format
+
+    def __repr__(self):
+        return f"@{self.namespace}{self.name}{self.path} ({self.file_format})"
+
+    @classmethod
+    def from_parent_stage(cls, parent_stage, path, file_format=None):
+        """
+        Extend an existing parent stage (with or without path) with an
+        additional sub-path
+        """
+        return cls(parent_stage.name,
+                   parent_stage.path + "/" + path,
+                   parent_stage.namespace,
+                   file_format)
+
+
+class CreateFileFormat(DDLElement):
+    """
+    Encapsulates a CREATE FILE FORMAT statement; using a format description (as in
+    a COPY INTO statement) and a format name.
+    """
+    __visit_name__ = "create_file_format"
+
+    def __init__(self, format_name, formatter, replace_if_exists=False):
+        super().__init__()
+        self.format_name = format_name
+        self.formatter = formatter
+        self.replace_if_exists = replace_if_exists
+
+
+class CreateStage(DDLElement):
+    """
+    Encapsulates a CREATE STAGE statement, using a container (physical base for the
+    stage) and the actual ExternalStage object.
+    """
+    __visit_name__ = "create_stage"
+
+    def __init__(self, container, stage, replace_if_exists=False):
+        super().__init__()
+        self.container = container
+        self.stage = stage
+        self.replace_if_exists = replace_if_exists
+
 
 class AWSBucket(ClauseElement):
     """AWS S3 bucket descriptor"""
@@ -360,7 +526,7 @@ class AzureContainer(ClauseElement):
         self.container = container
         self.path = path
         self.encryption_used = {}
-        self.credential_used = {}
+        self.credentials_used = {}
 
     @classmethod
     def from_uri(cls, uri):
@@ -389,7 +555,9 @@ class AzureContainer(ClauseElement):
             self.container,
             '/' + self.path if self.path else ""
         )
-        return uri + credentials if self.credentials_used else '' + encryption if self.encryption_used else ''
+        return '{}{}{}'.format(uri,
+                               f' {credentials}' if self.credentials_used else '',
+                               f' {encryption}' if self.encryption_used else '')
 
     def credentials(self, azure_sas_token):
         self.credentials_used = {'AZURE_SAS_TOKEN': azure_sas_token}
