@@ -5,30 +5,45 @@
 #
 
 import operator
-from six.moves.urllib_parse import unquote_plus
+from functools import reduce
 
 import sqlalchemy.types as sqltypes
 from six import iteritems
-from sqlalchemy import util as sa_util
+from six.moves.urllib_parse import unquote_plus
+from snowflake.connector import errors as sf_errors
+from snowflake.connector.constants import UTF8
 from sqlalchemy import event as sa_vnt
-from functools import reduce
 from sqlalchemy import exc as sa_exc
-from sqlalchemy.sql.elements import quoted_name
-from sqlalchemy.schema import Table
+from sqlalchemy import util as sa_util
 from sqlalchemy.engine import default, reflection
+from sqlalchemy.schema import Table
+from sqlalchemy.sql import text
+from sqlalchemy.sql.elements import quoted_name
 from sqlalchemy.types import (
-    CHAR, DATE, DATETIME, INTEGER, SMALLINT, BIGINT, DECIMAL, TIME, TIMESTAMP, VARCHAR, BINARY, BOOLEAN, FLOAT, REAL
+    BIGINT,
+    BINARY,
+    BOOLEAN,
+    CHAR,
+    DATE,
+    DATETIME,
+    DECIMAL,
+    FLOAT,
+    INTEGER,
+    REAL,
+    SMALLINT,
+    TIME,
+    TIMESTAMP,
+    VARCHAR,
 )
 
 from .base import (
-    SnowflakeDDLCompiler, SnowflakeCompiler, SnowflakeExecutionContext, SnowflakeIdentifierPreparer,
-    SnowflakeTypeCompiler
+    SnowflakeCompiler,
+    SnowflakeDDLCompiler,
+    SnowflakeExecutionContext,
+    SnowflakeIdentifierPreparer,
+    SnowflakeTypeCompiler,
 )
-from snowflake.connector.constants import UTF8
-from .custom_types import (
-    TIMESTAMP_LTZ, TIMESTAMP_TZ, TIMESTAMP_NTZ, VARIANT, OBJECT, ARRAY
-)
-from snowflake.connector import errors as sf_errors, ProgrammingError
+from .custom_types import ARRAY, OBJECT, TIMESTAMP_LTZ, TIMESTAMP_NTZ, TIMESTAMP_TZ, VARIANT
 
 colspecs = {}
 
@@ -188,8 +203,8 @@ class SnowflakeDialect(default.DefaultDialect):
         full_name = self._denormalize_quote_join(schema, object_name)
         try:
             results = connection.execute(
-                "DESC {0} /* sqlalchemy:_has_object */ {1}".format(
-                    object_type, full_name))
+                text("DESC {0} /* sqlalchemy:_has_object */ {1}".format(
+                    object_type, full_name)))
             row = results.fetchone()
             have = row is not None
             return have
@@ -227,10 +242,13 @@ class SnowflakeDialect(default.DefaultDialect):
 
     @reflection.cache
     def _current_database_schema(self, connection, **kw):
-        con = connection.connect().connection
-        return (
-            self.normalize_name(con.database),
-            self.normalize_name(con.schema))
+        # "branches" a connection if one exists, otherwise creates one
+        # branched connections use the same underlying DBAPI but maintain separate references
+        with connection.connect() as sqla_con:
+            dbapi_con = sqla_con.connection
+            return (
+                self.normalize_name(dbapi_con.database),
+                self.normalize_name(dbapi_con.schema))
 
     def _get_default_schema_name(self, connection):
         # NOTE: no cache object is passed here
@@ -254,9 +272,9 @@ class SnowflakeDialect(default.DefaultDialect):
 
     @reflection.cache
     def _get_schema_primary_keys(self, connection, schema, **kw):
-        result = connection.execute(
+        result = connection.execute(text(
             "SHOW /* sqlalchemy:_get_schema_primary_keys */PRIMARY KEYS IN SCHEMA {0}".format(schema)
-        )
+        ))
         ans = {}
         for row in result:
             table_name = self.normalize_name(row['table_name'])
@@ -279,9 +297,9 @@ class SnowflakeDialect(default.DefaultDialect):
 
     @reflection.cache
     def _get_schema_foreign_keys(self, connection, schema, **kw):
-        result = connection.execute(
+        result = connection.execute(text(
             "SHOW /* sqlalchemy:_get_schema_foreign_keys */ IMPORTED KEYS IN SCHEMA {0}".format(schema)
-        )
+        ))
         foreign_key_map = {}
         for row in result:
             name = self.normalize_name(row['fk_name'])
@@ -326,7 +344,7 @@ class SnowflakeDialect(default.DefaultDialect):
         full_schema_name = self._denormalize_quote_join(current_database, schema)
         try:
             schema_primary_keys = self._get_schema_primary_keys(connection, full_schema_name, **kw)
-            result = connection.execute("""
+            result = connection.execute(text("""
             SELECT /* sqlalchemy:_get_schema_columns */
                    ic.table_name,
                    ic.column_name,
@@ -339,8 +357,8 @@ class SnowflakeDialect(default.DefaultDialect):
                    ic.is_identity,
                    ic.comment
               FROM information_schema.columns ic
-             WHERE ic.table_schema=%(table_schema)s
-             ORDER BY ic.ordinal_position""", {"table_schema": self.denormalize_name(schema)})
+             WHERE ic.table_schema=:table_schema
+             ORDER BY ic.ordinal_position"""), {"table_schema": self.denormalize_name(schema)})
         except sa_exc.ProgrammingError as pe:
             if pe.orig.errno == 90030:
                 # This means that there are too many tables in the schema, we need to go more granular
@@ -402,7 +420,7 @@ class SnowflakeDialect(default.DefaultDialect):
         current_database, _ = self._current_database_schema(connection, **kw)
         full_schema_name = self._denormalize_quote_join(current_database, schema)
         schema_primary_keys = self._get_schema_primary_keys(connection, full_schema_name, **kw)
-        result = connection.execute("""
+        result = connection.execute(text("""
         SELECT /* sqlalchemy:get_table_columns */
                ic.table_name,
                ic.column_name,
@@ -415,9 +433,9 @@ class SnowflakeDialect(default.DefaultDialect):
                ic.is_identity,
                ic.comment
           FROM information_schema.columns ic
-         WHERE ic.table_schema=%(table_schema)s
-           AND ic.table_name=%(table_name)s
-         ORDER BY ic.ordinal_position""", {"table_schema": self.denormalize_name(schema),
+         WHERE ic.table_schema=:table_schema
+           AND ic.table_name=:table_name
+         ORDER BY ic.ordinal_position"""), {"table_schema": self.denormalize_name(schema),
                                            "table_name": self.denormalize_name(table_name)})
         for (table_name,
              column_name,
@@ -489,12 +507,12 @@ class SnowflakeDialect(default.DefaultDialect):
         schema = schema or self.default_schema_name
         current_schema = schema
         if schema:
-            cursor = connection.execute(
+            cursor = connection.execute(text(
                 "SHOW /* sqlalchemy:get_table_names */ TABLES IN {0}".format(
-                    self._denormalize_quote_join(schema)))
+                    self._denormalize_quote_join(schema))))
         else:
-            cursor = connection.execute(
-                "SHOW /* sqlalchemy:get_table_names */ TABLES")
+            cursor = connection.execute(text(
+                "SHOW /* sqlalchemy:get_table_names */ TABLES"))
             _, current_schema = self._current_database_schema(connection)
 
         ret = [self.normalize_name(row[1]) for row in cursor]
@@ -508,12 +526,12 @@ class SnowflakeDialect(default.DefaultDialect):
         """
         schema = schema or self.default_schema_name
         if schema:
-            cursor = connection.execute(
+            cursor = connection.execute(text(
                 "SHOW /* sqlalchemy:get_view_names */ VIEWS IN {0}".format(
-                    self._denormalize_quote_join((schema))))
+                    self._denormalize_quote_join((schema)))))
         else:
-            cursor = connection.execute(
-                "SHOW /* sqlalchemy:get_view_names */ VIEWS")
+            cursor = connection.execute(text(
+                "SHOW /* sqlalchemy:get_view_names */ VIEWS"))
 
         return [self.normalize_name(row[1]) for row in cursor]
 
@@ -524,16 +542,16 @@ class SnowflakeDialect(default.DefaultDialect):
         """
         schema = schema or self.default_schema_name
         if schema:
-            cursor = connection.execute(
+            cursor = connection.execute(text(
                 "SHOW /* sqlalchemy:get_view_definition */ VIEWS "
                 "LIKE '{0}' IN {1}".format(
                     self._denormalize_quote_join(view_name),
-                    self._denormalize_quote_join(schema)))
+                    self._denormalize_quote_join(schema))))
         else:
-            cursor = connection.execute(
+            cursor = connection.execute(text(
                 "SHOW /* sqlalchemy:get_view_definition */ VIEWS "
                 "LIKE '{0}'".format(
-                    self._denormalize_quote_join(view_name)))
+                    self._denormalize_quote_join(view_name))))
 
         n2i = self.__class__._map_name_to_idx(cursor)
         try:
@@ -547,13 +565,13 @@ class SnowflakeDialect(default.DefaultDialect):
     def get_temp_table_names(self, connection, schema=None, **kw):
         schema = schema or self.default_schema_name
         if schema:
-            cursor = connection.execute(
+            cursor = connection.execute(text(
                 "SHOW /* sqlalchemy:get_temp_table_names */ TABLES "
                 "IN {0}".format(
-                    self._denormalize_quote_join(schema)))
+                    self._denormalize_quote_join(schema))))
         else:
-            cursor = connection.execute(
-                "SHOW /* sqlalchemy:get_temp_table_names */ TABLES")
+            cursor = connection.execute(text(
+                "SHOW /* sqlalchemy:get_temp_table_names */ TABLES"))
 
         ret = []
         n2i = self.__class__._map_name_to_idx(cursor)
@@ -567,8 +585,8 @@ class SnowflakeDialect(default.DefaultDialect):
         """
         Gets all schema names.
         """
-        cursor = connection.execute(
-            "SHOW /* sqlalchemy:get_schema_names */ SCHEMAS")
+        cursor = connection.execute(text(
+            "SHOW /* sqlalchemy:get_schema_names */ SCHEMAS"))
 
         return [self.normalize_name(row[1]) for row in cursor]
 
@@ -581,7 +599,7 @@ class SnowflakeDialect(default.DefaultDialect):
                                         table_name,
                                         (' IN SCHEMA {}'.format(self.normalize_name(schema))) if schema else ''
                                     )
-        cursor = connection.execute(sql_command)
+        cursor = connection.execute(text(sql_command))
         return cursor.fetchone()
 
     def _get_view_comment(self, connection, table_name, schema=None, **kw):
@@ -593,7 +611,7 @@ class SnowflakeDialect(default.DefaultDialect):
                 table_name,
                 (' IN SCHEMA {}'.format(self.normalize_name(schema))) if schema else ''
             )
-        cursor = connection.execute(sql_command)
+        cursor = connection.execute(text(sql_command))
         return cursor.fetchone()
 
     def get_table_comment(self, connection, table_name, schema=None, **kw):
